package edu.uiuc.ncsa.myproxy.oa4mp.client.loader;

import edu.uiuc.ncsa.myproxy.oa4mp.client.ClientEnvironment;
import edu.uiuc.ncsa.myproxy.oa4mp.client.ClientLoaderInterface;
import edu.uiuc.ncsa.myproxy.oa4mp.client.ClientXMLTags;
import edu.uiuc.ncsa.myproxy.oa4mp.client.OA4MPServiceProvider;
import edu.uiuc.ncsa.myproxy.oa4mp.client.storage.*;
import edu.uiuc.ncsa.security.core.configuration.Configurations;
import edu.uiuc.ncsa.security.core.configuration.provider.CfgEvent;
import edu.uiuc.ncsa.security.core.configuration.provider.TypedProvider;
import edu.uiuc.ncsa.security.core.exceptions.ConfigurationException;
import edu.uiuc.ncsa.security.core.util.BasicIdentifier;
import edu.uiuc.ncsa.security.delegation.servlet.StorageConfigLoader;
import edu.uiuc.ncsa.security.delegation.storage.Client;
import edu.uiuc.ncsa.security.delegation.token.TokenForge;
import edu.uiuc.ncsa.security.oauth_1_0a.OAuthConstants;
import edu.uiuc.ncsa.security.oauth_1_0a.OAuthTokenForge;
import edu.uiuc.ncsa.security.oauth_1_0a.client.OAClient;
import edu.uiuc.ncsa.security.oauth_1_0a.client.OAuthDelegationServiceProvider;
import edu.uiuc.ncsa.security.util.pkcs.KeyUtil;
import net.oauth.OAuth;
import org.apache.commons.configuration.tree.ConfigurationNode;

import javax.inject.Provider;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.Date;
import java.util.HashMap;

import static edu.uiuc.ncsa.myproxy.oa4mp.client.ClientEnvironment.CALLBACK_URI_KEY;
import static edu.uiuc.ncsa.myproxy.oa4mp.client.ClientXMLTags.*;

/**
 * Referenced in the deployment descriptor, this class will process the configuration file and return
 * a {@link ClientEnvironment}.
 * <p>Created by Jeff Gaynor<br>
 * on 3/23/12 at  8:04 AM
 */
public class ClientLoader<T extends ClientEnvironment> extends StorageConfigLoader<T> implements ClientLoaderInterface {
    @Override
    public String getVersionNumber() {
        return "OA4MP Client version 1.0.6";
    }

    public ClientLoader(ConfigurationNode configurationNode) {
        super(configurationNode);
    }

    public static final String INITIATE_ENDPOINT = "initiate";
    public static final String ACCESS_TOKEN_ENDPOINT = "token";
    public static final String AUTHORIZE_ENDPOINT = "authorize";
    public static final String ASSET_ENDPOINT = "getcert";
    public static final long defaultCertLifetime = 0L;

    /**
     * Factory method. Override this to create the actual instance as needed.
     *
     * @param tokenForgeProvider
     * @param clientProvider
     * @param constants
     * @return
     */
    public T createInstance(Provider<TokenForge> tokenForgeProvider,
                            Provider<Client> clientProvider,
                            HashMap<String, String> constants) {
        ClientEnvironment ce = null;
        try {
            ce = new ClientEnvironment(
                    myLogger, constants,
                    getAccessTokenURI(),
                    getAuthorizeURI(),
                    getCallback(),
                    getInitiateURI(),
                    getAssetURI(),
                    checkCertLifetime(),
                    getId(),
                    checkPrivateKey(),
                    checkPublicKey(),
                    getSkin(),
                    isEnableAssetCleanup(),
                    getMaxAssetLifetime(),
                    clientProvider,
                    tokenForgeProvider,
                    getDSP(),
                    getAssetStoreProvider()
            );
        } catch (IOException e) {
            throw new ConfigurationException("Error; Could not configure environment", e);
        }
        return (T) ce;
    }


    Provider<AssetStore> assetStoreProvider;

    Provider<AssetStore> getAssetStoreProvider() {
        if (assetStoreProvider == null) {
            MultiAssetStoreProvider masp = new MultiAssetStoreProvider(cn, isDefaultStoreDisabled(), loggerProvider.get());
            final AssetProvider assetProvider = new AssetProvider();
            AssetConverter assetConverter = new AssetConverter(new AssetSerializationKeys(), assetProvider);
            assetStoreProvider = masp;
            masp.addListener(new FSAssetStoreProvider(cn, assetConverter));
            masp.addListener(new SQLAssetStoreProvider(cn, ClientXMLTags.POSTGRESQL_STORE, getPgConnectionPoolProvider(),
                    assetProvider, assetConverter));
            masp.addListener(new SQLAssetStoreProvider(cn, ClientXMLTags.MYSQL_STORE, getMySQLConnectionPoolProvider(),
                    assetProvider, assetConverter));
            // and a memory store, So only if one is requested it is available.
            masp.addListener(new TypedProvider<MemoryAssetStore>(cn, ClientXMLTags.MEMORY_STORE, ClientXMLTags.ASSET_STORE) {
                @Override
                public Object componentFound(CfgEvent configurationEvent) {
                    if (checkEvent(configurationEvent)) {
                        return get();
                    }
                    return null;
                }

                @Override
                public MemoryAssetStore get() {
                    return new MemoryAssetStore(assetProvider);
                }
            });
        }
        return assetStoreProvider;
    }

    OAuthDelegationServiceProvider dsp;

    protected OAuthDelegationServiceProvider getDSP() {
        if (dsp == null) {
            dsp = new OAuthDelegationServiceProvider(getInitiateURI(),
                    getAccessTokenURI(),
                    getAssetURI());

        }
        return dsp;
    }

    @Override
    public T createInstance() {

        Provider<TokenForge> tokenForgeProvider = new Provider<TokenForge>() {
            @Override
            public TokenForge get() {
                return new OAuthTokenForge(getId());
            }
        };

        Provider<Client> clientProvider = new Provider<Client>() {
            @Override
            public Client get() {
                OAClient c = new OAClient(BasicIdentifier.newID(getId()));
                c.setSignatureMethod(OAuthConstants.RSA_SHA1);
                c.setCreationTS(new Date());
                return c;
            }
        };

        // sets constants specific to this protocol.
        HashMap<String, String> constants = new HashMap<String, String>();
        constants.put(CALLBACK_URI_KEY, OAuthConstants.OAUTH_CALLBACK);
        constants.put(ClientEnvironment.FORM_ENCODING, "UTF-8");
        constants.put(ClientEnvironment.TOKEN, OAuth.OAUTH_TOKEN);

        return createInstance(tokenForgeProvider, clientProvider, constants);
    }

    /**
     * This takes a key and returns the value of the node associated with that key.
     *
     * @param key
     * @return
     */
    protected String getCfgValue(String key) {
        return Configurations.getNodeValue(cn, key);
    }

    T loader;

    @Override
    public T load() {
        if (loader == null) {
            loader = createInstance();
        }
        return loader;
    }

    protected PrivateKey checkPrivateKey() throws IOException {
        String privateKeyFileName = getCfgValue(ClientXMLTags.PRIVATE_KEY);
        if (trivial(privateKeyFileName)) {
            throw new ConfigurationException("Error: There is no private key specified.");
        }
        File privateKeyFile = new File(privateKeyFileName);
        if (!privateKeyFile.exists()) {
            throw new ConfigurationException("Error: The specified private key file \"" + privateKeyFileName + "\" does not exist");
        }
        if (!privateKeyFile.isFile()) {
            throw new ConfigurationException("Error: The specified private key file \"" + privateKeyFileName + "\" is not actually a file");
        }

        if (!privateKeyFile.canRead()) {
            throw new ConfigurationException("Error: The specified private key file \"" + privateKeyFileName + "\" is not readable. Check the permissions.");
        }

        return KeyUtil.fromPKCS8PEM(new FileReader(privateKeyFile));
    }

    protected PublicKey checkPublicKey() throws IOException {
        String publicKeyFileName = getCfgValue(ClientXMLTags.PUBLIC_KEY);
        if (trivial(publicKeyFileName)) {
            throw new ConfigurationException("Error: There is no public key specified.");
        }
        File publicKeyFile = new File(publicKeyFileName);
        if (!publicKeyFile.exists()) {
            throw new ConfigurationException("Error: The specified public key file \"" + publicKeyFileName + "\" does not exist");
        }
        if (!publicKeyFile.isFile()) {
            throw new ConfigurationException("Error: The specified public key file \"" + publicKeyFileName + "\" is not actually a file");
        }

        if (!publicKeyFile.canRead()) {
            throw new ConfigurationException("Error: The specified public key file \"" + publicKeyFileName + "\" is not readable. Check the permissions.");
        }

        return KeyUtil.fromX509PEM(new FileReader(publicKeyFile));
    }

    protected String getId() {
        String id = getCfgValue(ClientXMLTags.ID);
        if (trivial(id)) {
            throw new ConfigurationException("Error: there is no identifier specified.");
        }
        return id;
    }

    protected URI getCallback() {
        return checkURI(getCfgValue(ClientXMLTags.CALLBACK_URI), "callback");
    }


    protected URI checkURI(String x, String componentName) {
        if (trivial(x)) {
            throw new ConfigurationException("Error: There is no " + componentName + " URI specified.");
        }
        try {
            // set it this way rather than with URI.create so we get a recognizable exception to hand back.
            return new URI(x);
        } catch (URISyntaxException e) {
            throw new ConfigurationException("Error: The specified " + componentName + " is not a valid URI", e);
        }

    }

    protected URI getInitiateURI() {
        return createServiceURI(getCfgValue(ClientXMLTags.INITIATE_URI), getCfgValue(ClientXMLTags.BASE_URI), INITIATE_ENDPOINT);
    }

    protected URI getAccessTokenURI() {
        return createServiceURI(getCfgValue(ClientXMLTags.ACCESS_TOKEN_URI), getCfgValue(ClientXMLTags.BASE_URI), ACCESS_TOKEN_ENDPOINT);
    }

    protected URI getAssetURI() {
        return createServiceURI(getCfgValue(ClientXMLTags.ASSET_URI), getCfgValue(ClientXMLTags.BASE_URI), ASSET_ENDPOINT);
    }

    protected URI getAuthorizeURI() {
        return createServiceURI(getCfgValue(ClientXMLTags.AUTHORIZE_TOKEN_URI), getCfgValue(ClientXMLTags.BASE_URI), AUTHORIZE_ENDPOINT);
    }


    /**
     * Checks if there is a found uri in the configuration. If so, check that and use it,
     * if not, try and construct the service endpoint from the base uri.
     *
     * @param foundURI
     * @param baseUri
     * @param serviceEndpoint
     * @return
     */
    protected URI createServiceURI(String foundURI, String baseUri, String serviceEndpoint) {
        if (!trivial(foundURI)) {
            return checkURI(foundURI, serviceEndpoint);
        }
        if (trivial(baseUri)) {
            throw new ConfigurationException("Error: No base uri for " + serviceEndpoint + " found");
        }
        return checkURI(baseUri + (baseUri.endsWith("/") ? "" : "/") + serviceEndpoint, serviceEndpoint);
    }

    protected long checkCertLifetime() {
        String certLifetimeString = getCfgValue(ClientXMLTags.CERT_LIFETIME);
        if (!trivial(certLifetimeString)) {
            try {
                return Long.parseLong(certLifetimeString);
            } catch (Throwable t) {
                // if it fails, just say so, but keep going.
                myLogger.warn("Error: parsing default lifetime for cert:" + t.getMessage());
            }
        }
        return defaultCertLifetime;
    }


    boolean trivial(String x) {
        return x == null || 0 == x.length();
    }


    /**
     * This will return a service provider. Do not call this (or for that matter any other getter)
     * until {@link #load()} has been called.
     *
     * @return
     */
    public OA4MPServiceProvider getServiceProvider() {
        return new OA4MPServiceProvider(load());
    }

    protected String getSkin() {
        return getCfgValue(SKIN);
    }

    protected long getMaxAssetLifetime() {
        String x = getCfgValue(MAX_ASSET_LIFETIME);
        long y = 30 * 24 * 3600000L; // set it for a month in milliseconds
        if(x == null || x.length() == 0) return y;
        try {
            // The value in the file is assumed to be in seconds.
            y = Long.parseLong(x)*1000L;
        } catch (NumberFormatException nfx) {
            // use default...
        }
        return y;
    }

    protected boolean isEnableAssetCleanup(){
        boolean doIt = false;
        try{
            doIt = Boolean.parseBoolean(getCfgValue(ENABLE_ASSET_CLEANUP));
        }catch(Throwable t){
            // do nothing....
        }
        return doIt;
    }
}
